"use strict";
/**
 * Render
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Render = void 0;
const utils_1 = require("../utils/utils");
const config_1 = require("../config/config");
const Container_1 = require("../layouts/Container/Container");
const Column_1 = require("../layouts/Column/Column");
const Form_1 = require("../objects/Form/Form");
const Field_1 = require("../objects/Field/Field");
const RichText_1 = require("../text/RichText/RichText");
/**
 * Store slug data for json
 *
 * @private
 * @type {import('./RenderTypes').RenderSlugs}
 */
const _slugs = {};
/**
 * Function - normalize meta properties into one object
 *
 * @private
 * @param {import('./RenderTypes').RenderMetaArgs} args
 * @return {import('./RenderTypes').RenderMetaReturn}
 */
const _getMeta = (args) => {
    const meta = {
        title: '',
        description: '',
        url: '',
        image: '',
        canonical: '',
        prev: '',
        next: '',
        noIndex: false,
        isIndex: false
    };
    if (!(0, utils_1.isObjectStrict)(args)) {
        return meta;
    }
    if ((0, utils_1.isObjectStrict)(args.meta)) {
        return Object.assign(meta, args.meta);
    }
    if ((0, utils_1.isStringStrict)(args.metaTitle)) {
        meta.title = args.metaTitle;
    }
    if ((0, utils_1.isStringStrict)(args.metaDescription)) {
        meta.description = args.metaDescription;
    }
    const imageUrl = utils_1.getProp.file(args.metaImage, 'url');
    meta.image = (0, utils_1.isStringStrict)(imageUrl) ? imageUrl : '';
    return meta;
};
/**
 * Function - add pagination data to meta object
 *
 * @private
 * @param {import('../components/Pagination/PaginationTypes').PaginationData} args
 * @return {void}
 */
const _setPaginationMeta = (args, slugArgs, meta) => {
    const { current = 0, currentFilters, prevFilters, nextFilters } = args;
    slugArgs.page = current > 1 ? current : 0;
    const c = (0, utils_1.getSlug)(slugArgs);
    if ((0, utils_1.isObject)(c)) {
        meta.canonical = `${(0, utils_1.getPermalink)(c.slug, current === 1)}${(0, utils_1.isString)(currentFilters) ? currentFilters : ''}`;
    }
    if ((0, utils_1.isStringStrict)(args.title)) {
        meta.paginationTitle = args.title;
    }
    if ((0, utils_1.isNumber)(args.prev)) {
        slugArgs.page = args.prev > 1 ? args.prev : 0;
        const p = (0, utils_1.getSlug)(slugArgs);
        if ((0, utils_1.isObject)(p)) {
            meta.prev = `${(0, utils_1.getPermalink)(p.slug, args.prev === 1)}${(0, utils_1.isString)(prevFilters) ? prevFilters : ''}`;
        }
    }
    if ((0, utils_1.isNumber)(args.next) && args.next > 1) {
        slugArgs.page = args.next;
        const n = (0, utils_1.getSlug)(slugArgs);
        if ((0, utils_1.isObject)(n)) {
            meta.next = `${(0, utils_1.getPermalink)(n.slug, false)}${(0, utils_1.isString)(nextFilters) ? nextFilters : ''}`;
        }
    }
};
/**
 * Function - map out content slots to templates for contentTemplate
 *
 * @private
 * @param {object|object[]} templates
 * @param {object[]} content
 * @return {undefined}
 */
const _mapContentTemplate = (templates, content) => {
    if (!(0, utils_1.isArrayStrict)(content)) {
        return;
    }
    /* Transform templates */
    if ((0, utils_1.isArrayStrict)(templates)) {
        templates.forEach((t, i) => {
            /* Remove template break */
            if ((0, utils_1.isObjectStrict)(utils_1.getProp.tag(content[0], 'templateBreak'))) {
                content.shift();
            }
            /* Check if slot */
            const isSlot = (0, utils_1.isObjectStrict)(utils_1.getProp.tag(t, 'templateSlot'));
            /* Check for repeatable template item */
            const children = utils_1.getProp.fields(t, 'content');
            const repeat = utils_1.getProp.fields(t, 'repeat');
            if ((0, utils_1.isObjectStrict)(repeat) && (0, utils_1.isArrayStrict)(children) && !isSlot) {
                const repeatId = utils_1.getProp.id(repeat);
                const repeatIndex = children.findIndex((c) => {
                    return utils_1.getProp.id(c) === repeatId;
                });
                if (repeatIndex !== -1) {
                    let breakIndex = content.findIndex((c) => {
                        return (0, utils_1.isObjectStrict)(utils_1.getProp.tag(c, 'templateBreak'));
                    });
                    breakIndex = breakIndex === -1 ? content.length : breakIndex;
                    let insertIndex = repeatIndex;
                    for (let i = 0; i < breakIndex - 1; i += 1) {
                        children.splice(insertIndex, 0, structuredClone(repeat));
                        insertIndex += 1;
                    }
                }
            }
            /* Replace slot with content */
            if (isSlot) {
                templates.unshift();
                templates[i] = content.shift();
                return;
            }
            /* Recurse children */
            if ((0, utils_1.isObject)(children)) {
                _mapContentTemplate(children, content);
            }
        });
    }
    /* Check for nested content */
    if ((0, utils_1.isObjectStrict)(templates)) {
        (0, utils_1.getObjectKeys)(templates).forEach((t) => {
            const template = templates[t];
            const children = utils_1.getProp.fields(template, 'content');
            if ((0, utils_1.isObject)(children)) {
                _mapContentTemplate(children, content);
            }
        });
    }
};
/**
 * Function - recurse and output nested content
 *
 * @private
 * @param {import('./RenderTypes').RenderContentArgs} args
 * @return {void}
 */
const _renderContent = async (args) => {
    if (!(0, utils_1.isObject)(args)) {
        return;
    }
    const { contentData = [], serverlessData, output, pageData, pageContains = [], pageHeadings = [], navigations = {}, renderFunctions = {} } = args;
    let { parents = [], headingsIndex = 0, depth = 0 } = args;
    /* Content must be array */
    if (!(0, utils_1.isArrayStrict)(contentData)) {
        return;
    }
    /* Loop */
    for (let i = 0; i < contentData.length; i += 1) {
        let c = contentData[i];
        if (!(0, utils_1.isObjectStrict)(c)) {
            continue;
        }
        /* Store html string and filter info */
        let filterType = '';
        let filterArgs = {};
        let richTextOutput = '';
        /* Check for embedded entries and rich text */
        const richTextNodeType = c.nodeType;
        if ((0, utils_1.isStringStrict)(richTextNodeType)) {
            if (richTextNodeType === 'embedded-entry-block' || richTextNodeType === 'embedded-asset-block') {
                c = (0, utils_1.isObjectStrict)(c.data?.target) ? c.data.target : c;
            }
            else {
                let richTextContent;
                if ((0, utils_1.isStringStrict)(c.content) || (0, utils_1.isArrayStrict)(c.content)) {
                    richTextContent = c.content;
                }
                const richTextArgs = {
                    args: {
                        type: richTextNodeType,
                        content: richTextContent
                    },
                    parents,
                    headings: pageHeadings[headingsIndex]
                };
                richTextOutput = await (0, RichText_1.RichText)(richTextArgs);
                filterType = 'richText';
                filterArgs = {
                    type: richTextNodeType,
                    content: richTextContent
                };
            }
        }
        /* Render props */
        const normalProps = utils_1.getProp.self(c);
        const normalRenderType = utils_1.getProp.type(c, 'render');
        const props = (0, utils_1.isObjectStrict)(normalProps) ? normalProps : {};
        const renderType = (0, utils_1.isString)(normalRenderType) ? normalRenderType : '';
        const id = utils_1.getProp.id(c);
        if ((0, utils_1.isStringStrict)(id)) {
            props.id = id;
        }
        /* Check for nested content */
        let children = props.content;
        let recurse = children !== undefined;
        if ((0, utils_1.isObjectStrict)(children)) {
            const nestedChildren = children.content;
            if (children.nodeType !== undefined && nestedChildren !== undefined) {
                children = nestedChildren;
                recurse = true;
            }
        }
        /* Map out content to template */
        if (renderType === 'contentTemplate') {
            const templates = (0, utils_1.isArray)(props.templates) ? props.templates : [];
            _mapContentTemplate(templates, props.content);
            children = templates;
            recurse = true;
        }
        /* Render output */
        let renderObj = {
            start: '',
            end: ''
        };
        const renderFunction = renderFunctions[renderType];
        if ((0, utils_1.isFunction)(renderFunction)) {
            const renderArgs = {
                args: props,
                parents,
                pageData,
                pageContains,
                navigations,
                serverlessData
            };
            const renderOutput = await renderFunction(renderArgs);
            if ((0, utils_1.isString)(renderOutput)) {
                renderObj.start = renderOutput;
            }
            else {
                renderObj = renderOutput;
            }
            pageContains.push(renderType);
            filterType = renderType;
            filterArgs = props;
        }
        let start = renderObj.start;
        let end = renderObj.end;
        /* Filter start and end output */
        const renderContentFilterArgs = {
            renderType: filterType,
            args: filterArgs
        };
        if (start !== '' && end === '') {
            start = await (0, utils_1.applyFilters)('renderContent', start, renderContentFilterArgs);
        }
        else {
            start = await (0, utils_1.applyFilters)('renderContentStart', start, renderContentFilterArgs);
            end = await (0, utils_1.applyFilters)('renderContentEnd', end, renderContentFilterArgs);
        }
        if (richTextOutput !== '') {
            richTextOutput = await (0, utils_1.applyFilters)('renderContent', richTextOutput, renderContentFilterArgs);
        }
        /* Add to output object */
        output.html += richTextOutput;
        output.html += start;
        /* Recurse through children */
        if ((0, utils_1.isArrayStrict)(children) && recurse) {
            const parentsCopy = [...parents];
            parentsCopy.unshift({
                renderType,
                args: props
            });
            await _renderContent({
                contentData: children,
                serverlessData,
                output,
                parents: parentsCopy,
                pageData,
                pageContains,
                pageHeadings,
                navigations,
                renderFunctions,
                headingsIndex,
                depth: depth += 1
            });
        }
        output.html += end;
        /* Additional rich text areas */
        if (renderType === 'content' && depth === 0) {
            headingsIndex = pageHeadings.push([]);
        }
        /* Clear parents */
        if (renderType !== '' && renderType !== 'content' && end !== '') {
            parents = [];
        }
    }
};
/**
 * Function - output single post or page
 *
 * @private
 * @param {import('./RenderTypes').RenderItemArgs} args
 * @return {import('./RenderTypes').RenderItemReturn}
 */
const _renderItem = async (args) => {
    if (!(0, utils_1.isObjectStrict)(args)) {
        return {};
    }
    const { item, contentType = 'page', serverlessData, renderFunctions } = args;
    /* Item must be object */
    if (!(0, utils_1.isObjectStrict)(item)) {
        return {};
    }
    /* Item id required */
    const id = utils_1.getProp.id(item);
    if (!(0, utils_1.isStringStrict)(id)) {
        return {};
    }
    /* Item props */
    const props = utils_1.getProp.self(item);
    if (!(0, utils_1.isObjectStrict)(props)) {
        return {};
    }
    /* Slug required */
    if (!(0, utils_1.isStringStrict)(props.slug)) {
        return {};
    }
    /* Serverless render check */
    let serverlessRender = false;
    /* Store components contained in page  */
    const pageContains = [];
    /* Store rich text headings in page */
    const pageHeadings = [[]];
    /* Start action */
    const renderItemStartArgs = {
        id,
        pageData: props,
        contentType,
        pageContains,
        pageHeadings,
        serverlessData
    };
    await (0, utils_1.doActions)('renderItemStart', renderItemStartArgs);
    /* Reset script and style files */
    config_1.config.scripts.item = {};
    config_1.config.styles.item = {};
    /* Meta */
    const title = props.title;
    const meta = _getMeta(props);
    if (!(0, utils_1.isStringStrict)(meta.title)) {
        meta.title = title;
    }
    /* Permalink */
    const slugArgs = {
        id,
        contentType,
        linkContentType: props.linkContentType,
        slug: props.slug,
        returnParents: true,
        page: 0
    };
    const s = (0, utils_1.getSlug)(slugArgs);
    let slug = '';
    let permalink = '';
    let parents = [];
    if (!(0, utils_1.isString)(s)) {
        slug = s.slug;
        parents = s.parents;
        permalink = (0, utils_1.getPermalink)(slug);
        item.basePermalink = (0, utils_1.getPermalink)(slug);
    }
    meta.url = permalink;
    meta.canonical = permalink;
    /* Add to data by slugs store */
    const formattedSlug = slug !== 'index' && slug !== '' ? `/${slug}/` : '/';
    _slugs[formattedSlug] = {
        contentType,
        id
    };
    /* Check if index */
    const index = props.slug === 'index';
    meta.isIndex = index;
    /* Navigations */
    let navigations = {};
    if ((0, utils_1.isFunction)(renderFunctions.navigations)) {
        navigations = renderFunctions.navigations({
            navigations: config_1.config.navigation,
            items: config_1.config.navigationItem,
            current: permalink,
            title,
            parents
        });
    }
    /* Serverless data */
    let itemServerlessData;
    if ((0, utils_1.isObjectStrict)(serverlessData)) {
        const serverlessPath = serverlessData.path !== undefined ? serverlessData.path : '';
        if (serverlessPath === formattedSlug && serverlessData.query !== undefined) {
            itemServerlessData = serverlessData;
        }
        else { // Avoid re-rendering non dynamic pages
            return {
                serverlessRender: false
            };
        }
    }
    /* Content loop */
    const contentOutput = { html: '' };
    let contentData = props.content;
    if ((0, utils_1.isObjectStrict)(contentData)) {
        contentData = contentData.content;
    }
    if ((0, utils_1.isArrayStrict)(contentData)) {
        await _renderContent({
            contentData,
            serverlessData: itemServerlessData,
            output: contentOutput,
            parents: [],
            pageData: item,
            pageContains,
            pageHeadings,
            navigations,
            renderFunctions
        });
    }
    contentOutput.html = await (0, utils_1.doShortcodes)(contentOutput.html);
    /* Pagination variables for meta object */
    if ((0, utils_1.isObjectStrict)(item.pagination)) {
        _setPaginationMeta(item.pagination, slugArgs, meta);
        serverlessRender = true;
    }
    /* Page data (props) for layout and actions */
    const pageData = { ...props };
    pageData.id = id;
    pageData.parents = parents;
    pageData.content = [];
    /* Output */
    let layoutOutput = '';
    if ((0, utils_1.isFunction)(renderFunctions.layout)) {
        const layoutArgs = {
            id,
            meta,
            navigations,
            contentType,
            content: contentOutput.html,
            slug: formattedSlug,
            pageContains,
            pageHeadings,
            pageData,
            serverlessData
        };
        layoutOutput = await renderFunctions.layout(layoutArgs);
    }
    const renderItemFilterArgs = {
        id,
        contentType,
        slug: formattedSlug,
        output: layoutOutput,
        pageData,
        pageContains,
        pageHeadings,
        serverlessData
    };
    layoutOutput = await (0, utils_1.applyFilters)('renderItem', layoutOutput, renderItemFilterArgs);
    /* End action */
    const renderItemEndArgs = {
        id,
        contentType,
        slug: formattedSlug,
        output: layoutOutput,
        pageData,
        pageContains,
        pageHeadings,
        serverlessData
    };
    await (0, utils_1.doActions)('renderItemEnd', renderItemEndArgs);
    /* Output */
    return {
        serverlessRender,
        pageData,
        data: {
            slug: formattedSlug,
            output: layoutOutput
        }
    };
};
/**
 * Function - loop through all content types to output pages and posts
 *
 * @param {import('./RenderTypes').RenderArgs} args
 * @return {Promise<import('./RenderTypes').RenderReturn[]|import('./RenderTypes').RenderReturn>}
 */
const Render = async (args) => {
    /* Fallback output */
    const fallback = [{
            slug: '',
            output: ''
        }];
    /* Props must be object */
    if (!(0, utils_1.isObjectStrict)(args)) {
        return fallback;
    }
    const { allData, serverlessData, previewData } = args;
    /* Start action */
    await (0, utils_1.doActions)('renderStart', args);
    /* Reset script and style directories */
    config_1.config.scripts.build = {};
    config_1.config.styles.build = {};
    /* Data */
    if (!(0, utils_1.isObjectStrict)(allData)) {
        return fallback;
    }
    const { navigation, navigationItem, redirect, content } = allData;
    /* Store navigations and items */
    config_1.config.navigation = navigation;
    config_1.config.navigationItem = navigationItem;
    /* Add formation functions */
    const renderFunctions = { ...config_1.config.renderFunctions };
    if (renderFunctions.container === undefined) {
        renderFunctions.container = Container_1.Container;
    }
    if (renderFunctions.column === undefined) {
        renderFunctions.column = Column_1.Column;
    }
    if (renderFunctions.form === undefined) {
        renderFunctions.form = Form_1.Form;
    }
    if (renderFunctions.field === undefined) {
        renderFunctions.field = Field_1.Field;
    }
    if (renderFunctions.richText === undefined) {
        renderFunctions.richText = RichText_1.RichText;
    }
    /* Store content data */
    const data = [];
    /* Loop through pages first to set parent slugs */
    if (serverlessData === undefined) {
        for (let i = 0; i < content.page.length; i += 1) {
            const item = content.page[i];
            const itemProp = utils_1.getProp.self(item);
            const itemData = Object.assign({
                parent: undefined,
                archive: '',
                linkContentType: 'default'
            }, (0, utils_1.isObjectStrict)(itemProp) ? itemProp : {});
            let { parent, archive = '', linkContentType = 'default' } = itemData;
            archive = await (0, utils_1.applyFilters)('renderArchiveName', archive);
            linkContentType = await (0, utils_1.applyFilters)('renderLinkContentTypeName', linkContentType);
            const idProp = utils_1.getProp.id(item);
            const id = (0, utils_1.isStringStrict)(idProp) ? idProp : '';
            const isArchive = (0, utils_1.isStringStrict)(archive) && id !== '';
            if (isArchive && id !== '') {
                const archiveIds = config_1.config.archive.ids;
                const archiveSlug = itemProp.slug;
                const archiveTitle = itemProp.title;
                const contentTypeArchive = config_1.config.contentTypes.archive?.[archive]?.id;
                if (archiveIds[archive] === undefined) {
                    archiveIds[archive] = {};
                }
                archiveIds[archive][linkContentType] = id;
                if ((0, utils_1.isObjectStrict)(contentTypeArchive)) {
                    contentTypeArchive[linkContentType] = id;
                }
                if ((0, utils_1.isStringStrict)(archiveSlug) && (0, utils_1.isStringStrict)(archiveTitle)) {
                    config_1.config.slug.archives[id] = {
                        slug: archiveSlug,
                        title: archiveTitle,
                        contentType: archive
                    };
                }
            }
            if ((0, utils_1.isObjectStrict)(parent) && id !== '') {
                const parentSlug = utils_1.getProp.fields(parent, 'slug');
                const parentTitle = utils_1.getProp.fields(parent, 'title');
                const parentId = utils_1.getProp.id(parent);
                if ((0, utils_1.isStringStrict)(parentSlug) && (0, utils_1.isStringStrict)(parentTitle) && (0, utils_1.isStringStrict)(parentId)) {
                    config_1.config.slug.parents[id] = {
                        id: parentId,
                        slug: parentSlug,
                        title: parentTitle,
                        contentType: 'page'
                    };
                }
            }
        }
        if ((0, utils_1.isArrayStrict)(redirect)) {
            redirect.forEach((r) => {
                const redirectProp = utils_1.getProp.self(r);
                if (!(0, utils_1.isObjectStrict)(redirectProp)) {
                    return;
                }
                const { redirect: rr = [] } = redirectProp;
                if (!(0, utils_1.isArrayStrict)(rr)) {
                    return;
                }
                config_1.config.redirects.data = config_1.config.redirects.data.concat(rr);
            });
        }
    }
    else {
        const slugParentsData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('slugParents', 'store'));
        const slugArchivesData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('slugArchives', 'store'));
        const archiveIdsData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('archiveIds', 'store'));
        const archiveTermsData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('archiveTerms', 'store'));
        const archivePostsData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('archivePosts', 'store'));
        const formMetaData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('formMeta', 'store'));
        const navigationsData = await (0, utils_1.getJsonFile)((0, utils_1.getPath)('navigations', 'store'));
        if (slugParentsData !== undefined) {
            Object.keys(slugParentsData).forEach((s) => {
                config_1.config.slug.parents[s] = slugParentsData[s];
            });
        }
        if (slugArchivesData !== undefined) {
            Object.keys(slugArchivesData).forEach((s) => {
                config_1.config.slug.archives[s] = slugArchivesData[s];
            });
        }
        if (archiveIdsData !== undefined) {
            config_1.config.archive.ids = archiveIdsData;
            Object.keys(archiveIdsData).forEach((a) => {
                if (config_1.config.contentTypes.archive[a] != null) {
                    config_1.config.contentTypes.archive[a].id = archiveIdsData[a];
                }
            });
        }
        if (archiveTermsData !== undefined) {
            config_1.config.archive.terms = archiveTermsData;
        }
        if (archivePostsData !== undefined) {
            config_1.config.archive.posts = archivePostsData;
        }
        if (formMetaData !== undefined) {
            config_1.config.formMeta = formMetaData;
        }
        if (navigationsData !== undefined) {
            config_1.config.navigation = navigationsData.navigation;
            config_1.config.navigationItem = navigationsData.navigationItem;
        }
    }
    /* Empty serverless reload */
    Object.keys(config_1.config.serverless.routes).forEach((r) => {
        config_1.config.serverless.routes[r] = [];
    });
    /* Loop through all content types */
    const contentTypes = Object.keys(content);
    for (let c = 0; c < contentTypes.length; c += 1) {
        const contentType = contentTypes[c];
        for (let i = 0; i < content[contentType].length; i += 1) {
            const item = await _renderItem({
                item: content[contentType][i],
                contentType,
                serverlessData,
                renderFunctions
            });
            const { serverlessRender = false, data: itemData } = item;
            if (itemData !== undefined) {
                data.push(itemData);
                if (serverlessRender && serverlessData === undefined) {
                    config_1.config.serverless.routes.reload.push({
                        path: itemData.slug.replace(/^\/|\/$/gm, '')
                    });
                }
            }
        }
    }
    /* Files data */
    if (serverlessData === undefined && previewData === undefined) {
        config_1.config.store.files.slugs.data = JSON.stringify(_slugs);
        config_1.config.store.files.slugParents.data = JSON.stringify(config_1.config.slug.parents);
        config_1.config.store.files.slugArchives.data = JSON.stringify(config_1.config.slug.archives);
        config_1.config.store.files.archiveIds.data = JSON.stringify(config_1.config.archive.ids);
        config_1.config.store.files.archiveTerms.data = JSON.stringify(config_1.config.archive.terms);
        config_1.config.store.files.archivePosts.data = JSON.stringify(config_1.config.archive.posts);
        config_1.config.store.files.formMeta.data = JSON.stringify(config_1.config.formMeta);
        config_1.config.store.files.navigations.data = JSON.stringify({
            navigations: config_1.config.navigation,
            navigationItems: config_1.config.navigationItem
        });
    }
    /* Output */
    const output = serverlessData !== undefined || previewData !== undefined ? data[0] : data;
    await (0, utils_1.doActions)('renderEnd', { ...args, data: output });
    return output;
};
exports.Render = Render;
